import pyPPG

import matplotlib.pyplot as plt
import scipy.io
import numpy as np
import pandas as pd
from dotmap import DotMap
from tkinter import filedialog
import mne
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
import os
import tkinter as tk
from tkinter import simpledialog
from scipy.io import savemat

###########################################################################
####################### Data Acquisition from Files #######################
###########################################################################
def load_data(data_path = "", fs = np.nan, start_sig = 0, end_sig = -1, channel='Pleth', use_tk=True, print_flag=True):
    """
    Load raw PPG data.

    :param data_path: path of the file containing the PPG signal
    :type data_path: str
    :param start_sig: the first sample of the signal to be analysed
    :type start_sig: int
    :param fs: the sampling frequency of the PPG in Hz
    :type fs: int
    :param end_sig: the last sample of the signal to be analysed
    :type end_sig: int
    :param channel: channel of the .edf file
    :type channel: channel of the .edf file
    :param use_tk: a bool for using tkinter interface
    :type use_tk: bool
    :param print_flag: a bool for print message
    :type print_flag: bool

    :return: s: dictionary of the PPG signal:

        * s.start_sig: the first sample of the signal to be analysed
        * s.end_sig: the last sample of the signal to be analysed
        * s.v: a vector of PPG values
        * s.fs: the sampling frequency of the PPG in Hz
        * s.name: name of the record
        * s.v: 1-d array, a vector of PPG values
        * s.fs: the sampling frequency of the PPG in Hz
        * s.ppg: 1-d array, a vector of the filtered PPG values
        * s.vpg: 1-d array, a vector of the filtered PPG' values
        * s.apg: 1-d array, a vector of the filtered PPG" values
        * s.jpg: 1-d array, a vector of the filtered PPG'" values
        * s.filtering: a bool for filtering
        * s.correct: a bool for correcting fiducial points
    """

    if data_path=="":
        sig_path = filedialog.askopenfilename(title='Select SIGNAL file', filetypes=[("Input Files", ".mat .csv .edf .pkl .txt")])
    else:
        sig_path=data_path

    if sig_path.rfind('/')>0:
        start_c = sig_path.rfind('/')+1
    else:
        start_c = sig_path.rfind('\\')

    stop_c=sig_path.rfind('.')
    rec_name=sig_path[start_c:stop_c]

    try:
        sig_format=sig_path[len(sig_path)-sig_path[::-1].index('.'):]
    except:
        raise('Invalid signal path!')

    if sig_format=='mat':
        input_sig = scipy.io.loadmat(sig_path)
        sig = np.float64(np.squeeze(input_sig.get("Data")))[0:]
        try:
            fs = np.squeeze(input_sig.get("Fs"))
        except:
            fs = 125
            if print_flag: print('The default sampling frequency is 125 Hz for .mat.')
    elif sig_format=='csv':
        input_sig = pd.read_csv(sig_path, encoding='utf-8')
        sig = input_sig
        sig = np.squeeze(sig.values)

        if fs<=0:
            fs = 125
            if print_flag: print('The default sampling frequency is 125 Hz for .csv.')

    elif sig_format=='txt':
        try:
            input_sig = np.loadtxt(sig_path, delimiter='\t')
        except:
            try:
                input_sig = np.loadtxt(sig_path, delimiter=' ')
            except:
                print('ERROR! The data separator is not supported for .txt.')
        sig = input_sig

        if fs<=0:
            fs = 125
            if print_flag: print('The default sampling frequency is 125 Hz for .txt.')

    elif sig_format == 'edf':
        try:
            input_sig = mne.io.read_raw_edf(sig_path, include=channel)
            sig = -input_sig.get_data().squeeze()
        except:
            try:
                if use_tk: input_name = simpledialog.askstring("Input", "Define the PPG channel name:")
                input_sig = mne.io.read_raw_edf(sig_path, include=input_name)
                sig = -input_sig.get_data().squeeze()
            except:
                raise('There is no valid channel for PPG in the .edf file!')

        if fs<=0:
            try:
                fs = int(np.round(input_sig.info['sfreq']))
            except:
                fs = 256
                if print_flag: print('The default sampling frequency is 256 Hz for .edf.')

    s = DotMap()

    s.start_sig = start_sig
    if start_sig<end_sig:
        s.end_sig = end_sig
    else:
        s.end_sig = len(sig)

    try:
        s.v=sig[s.start_sig:s.end_sig]
    except:
        raise('There is no valid PPG signal!')

    s.fs=fs
    s.name=rec_name

    return s

###########################################################################
########################### Plot Fiducial points ##########################
###########################################################################
def plot_fiducials(s: pyPPG.PPG, fp: pyPPG.Fiducials, savefig=True, savingfolder='temp_dir', show_fig = True,
                   print_flag=True, use_tk=False, new_fig=True, marker=[], title='Detection', legend_loc='upper right',
                   legend_fontsize=20, marker_size=60, facecolor=False, subtext={}, canvas=np.nan):
    """
    Plot fiducial points of the filtered PPG signal.

    :param s: object of PPG signal
    :type s: pyPPG.PPG object
    :param fp: object of fiducial points
    :type fp: pyPPG.Fiducials object
    :param savefig: a bool for save figure
    :type savefig: bool
    :param savingfolder: location of the saved figure
    :param show_fig: a bool for show figure
    :type show_fig: bool
    :param print_flag: a bool for print message
    :type print_flag: bool
    :param use_tk: a bool for using tkinter interface
    :type use_tk: bool
    :param new_fig: a bool for creating new figure
    :type new_fig: bool
    :param marker: list of fiducial points markers
    :type marker: list
    :param title: title of the legend
    :type title: str
    :param legend_loc: location of the legend
    :type legend_loc: str
    :param legend_fontsize: fontsize of the legends
    :type legend_fontsize: int
    :param marker_size: size of markers
    :type marker_size: int
    :param facecolor: a bool for facecolor of markers
    :type facecolor: bool
    :param subtext: dictionary for subplots text
    :type subtext: dict
    :param canvas: canvas of the figure
    :type canvas: FigureCanvas
    """

    # Create a hidden root window to get screen dimensions
    if use_tk:
        root = tk.Tk()
        screen_width = root.winfo_screenwidth()
        screen_height = root.winfo_screenheight()
        root.destroy()
    else:
        screen_width = 1500
        screen_height = 950


    # Define a scaling factor for the figure size (e.g., 0.8 for 80% of the screen size)
    scaling_factor = 0.8

    # Calculate the figure size based on the screen dimensions and scaling factor
    figure_width = screen_width * scaling_factor
    figure_height = screen_height * scaling_factor

    if new_fig: fig = plt.figure(figsize=(figure_width/100, figure_height/100))
    ax1 = plt.subplot(411)
    plt.plot(s.ppg, 'k', label=None)
    ax2 = plt.subplot(412, sharex=ax1)
    plt.plot(s.vpg, 'k', label=None)
    ax3 = plt.subplot(413, sharex=ax2)
    plt.plot(s.apg, 'k', label=None)
    ax4 = plt.subplot(414, sharex=ax3)
    plt.plot(s.jpg, 'k', label=None)
    if new_fig:
        fig.subplots_adjust(hspace=0, wspace=0)
        canvas = FigureCanvas(fig)

    if len(marker)==0:
        marker = ['o', 's', 's','o', 'o', 's', 'o', 'o', 's', 'o', 's', 'o', 's', 'o', 's']

    color = ['r', 'b', 'g','m', 'r', 'b', 'g', 'r', 'b', 'g', 'm', 'c', 'k', 'r', 'b']

    fid_names = ('sp', 'on', 'dn','dp', 'u', 'v', 'w', 'a', 'b', 'c', 'd', 'e', 'f', 'p1', 'p2')
    ylabe_names = ['PPG', 'PPG\'', 'PPG\'\'', 'PPG\'\'\'']
    s_type = ['ppg', 'ppg', 'ppg','ppg', 'vpg', 'vpg', 'vpg', 'apg', 'apg', 'apg', 'apg', 'apg', 'apg', 'jpg','jpg']

    str_sig = 0
    end_sig = len(s.ppg)
    len_sig=end_sig-str_sig
    step_small = 1
    if len_sig/s.fs<10:
        step_big = step_small * 1
    else:
        step_big = step_small * 5

    major_ticks_names = range(0, int(len_sig/s.fs),step_big)
    len_ticks_names=len(major_ticks_names)
    major_diff=len_sig/len_ticks_names
    minor_diff = len_sig / len_ticks_names / step_big
    major_ticks = np.arange(str_sig, end_sig, major_diff)
    minor_ticks = np.arange(str_sig, end_sig, minor_diff)

    sig_names=('ppg','vpg','apg','jpg')
    for n in fid_names:
        ind = fid_names.index(n)
        tmp_ind = sig_names.index(s_type[ind])
        if tmp_ind == 0:
            plt_num = 1
            plt.subplot(411)
            plt.title(s.name, fontsize=20)
        else:
            plt_num = tmp_ind+1

        ax = plt.subplot(4, 1, plt_num)

        tmp_pnt=eval("fp." + n + ".values")
        tmp_pnt=tmp_pnt[~np.isnan(tmp_pnt)].astype(int)
        tmp_sig=eval("s." + s_type[ind])

        if facecolor:
            fc=color[ind]
        else:
            fc='none'

        exec("plt.scatter(tmp_pnt, tmp_sig[tmp_pnt], s=marker_size,linewidth=2, marker = marker[ind], facecolors=fc, color=color[ind], label=n)")

        if ind<len(s_type)-1:
            legend_flag = s_type[ind] != s_type[ind + 1]
        else:
            legend_flag=True

        if legend_flag:
            leg = ax.legend(loc=legend_loc, fontsize=legend_fontsize, ncol=2, facecolor="orange", frameon=True, title=title)
            ax.add_artist(leg)

            try:
                tmp_txt=subtext[s_type[ind]]
                plt.text(-.14, 0.95, tmp_txt, fontsize=legend_fontsize,transform=ax.transAxes, va='top', ha='left',
                         bbox={'facecolor': 'yellow', 'alpha': 0.5, 'edgecolor': 'black', 'pad': 5})
            except:
                pass

            if s_type[ind]=='jpg':
                try:
                    tmp_txt = 'MAE(SD):'+'\n'+str(subtext['mae'])+'('+str(subtext['std'])+')'
                    plt.text(-.14, 0, tmp_txt, fontsize=legend_fontsize, transform=ax.transAxes, va='top', ha='left', weight='bold', color='w',
                             bbox={'facecolor': 'red', 'alpha': 0.5, 'edgecolor': 'black', 'pad': 5})
                except:
                    pass

            plt.ylabel(ylabe_names[plt_num - 1], fontsize=20)

            exec("plt.xlim([str_sig,end_sig])")
            for text in leg.get_texts():
                text.set_weight('bold')

            ax.set_xticks(major_ticks)
            ax.set_xticks(minor_ticks, minor=True)

            ax.grid(which='both')
            ax.grid(which='minor', alpha=0.2,axis='x')
            ax.grid(which='major', alpha=1,axis='x')

            plt.yticks([])

    plt.xlabel('Time [s]', fontsize=20)
    plt.xticks(major_ticks,major_ticks_names, fontsize=20)
    if show_fig: plt.show()

    if not(':' in savingfolder):
        relative_path=r'.'+os.sep
    else:
        relative_path=''

    if savefig:
        tmp_dir=savingfolder+os.sep+'PPG_Figures'+os.sep

        os.makedirs(tmp_dir, exist_ok=True)

        canvas.print_png((tmp_dir+'%s_btwn_%s-%s.png') % (s.name,s.start_sig,s.end_sig))
        if print_flag: print('Figure has been saved in the "'+tmp_dir+'".')

    return canvas

###########################################################################
################################# Save Data ###############################
###########################################################################

def save_data(savingformat: str, savingfolder: str, print_flag=True, s={}, fp=pd.DataFrame(), bm=pd.DataFrame()):
    """
    Save the results of the filtered PPG analysis.

    :param savingformat: file format of the saved date, the provided file formats .mat and .csv
    :type savingformat: str
    :param savingfolder: location of the saved data
    :type savingfolder: str
    :param print_flag: a bool for print message
    :type print_flag: bool
    :param s: a struct of PPG signal
    :type s: pyPPG.PPG object
    :type s: DotMap
    :param fp: a struct of fiducial points
    :type fp: pyPPG.Fiducial object
    :param bm: a dictionary of biomarkers
    :type bm: pyPPG.Biomarkers object

    :return: file_names: dictionary of the saved file names
    """

    savingfolder = savingfolder.replace('/', '\\')

    if not(':' in savingfolder):
        relative_path=r'.'+os.sep
    else:
        relative_path=''

    tmp_dir = savingfolder
    os.makedirs(tmp_dir, exist_ok=True)

    temp_dirs = ['Fiducial_points', 'Biomarker_vals', 'Biomarker_stats', 'Biomarker_defs', 'PPG_struct', 'Biomarker_defs_and_stats']
    for i in temp_dirs:
        temp_dir = tmp_dir + os.sep + i + os.sep
        os.makedirs(temp_dir, exist_ok=True)

    keys=s.__dict__.keys()
    keys_list = list(keys)
    sc=DotMap()
    for i in keys_list:
        exec('sc.'+i+' = s.'+i)

    file_names = {}
    file_name = (relative_path + tmp_dir + os.sep + temp_dirs[4] + os.sep + s.name + '_data_btwn_%s-%s.mat')%(s.start_sig,s.end_sig)
    file_names ['data_struct_mat']= file_name
    scipy.io.savemat(file_name, sc)

    try:
        BM_keys = bm.bm_vals.keys()
    except:
        BM_keys={}


    if savingformat=="csv" or savingformat=="both":
        file_name = (relative_path+tmp_dir+os.sep+temp_dirs[0]+os.sep+s.name+'_'+'Fiducials_btwn_%s-%s.csv')%(s.start_sig,s.end_sig)
        file_names ['fiducials_csv']= file_name
        tmp_fp = fp.get_fp()
        tmp_fp.index = tmp_fp.index + 1
        tmp_fp.to_csv(file_name)

        for key in BM_keys:
            file_name = (relative_path+tmp_dir+os.sep+temp_dirs[1]+os.sep+'%s_btwn_%s-%s.csv')%(s.name+'_'+key,s.start_sig,s.end_sig)
            file_names [key + '_vals_csv']= file_name
            bm.bm_vals[key].index = bm.bm_vals[key].index + 1
            bm.bm_vals[key].to_csv(file_name,index=True,header=True)

            file_name = (relative_path+tmp_dir+os.sep+temp_dirs[2]+os.sep+'%s_btwn_%s-%s.csv')%(s.name+'_'+key,s.start_sig,s.end_sig)
            file_names [key + '_stats_csv']= file_name
            bm.bm_stats[key].to_csv(file_name, index=True, header=True)

            file_name = (relative_path+tmp_dir+os.sep+temp_dirs[3]+os.sep+'%s_btwn_%s-%s.csv')%(s.name+'_'+key,s.start_sig,s.end_sig)
            file_names [key + '_defs_csv']= file_name
            bm.bm_defs[key].index = bm.bm_defs[key].index + 1
            bm.bm_defs[key].to_csv(file_name, index=True, header=True)

    if savingformat=="mat"  or savingformat=="both":
        file_name = (relative_path+tmp_dir+os.sep+temp_dirs[0]+os.sep+s.name+'_'+'Fiducials_btwn_%s-%s.mat')%(s.start_sig,s.end_sig)
        file_names['fiducials_mat']=file_name
        tmp_fp = fp.get_fp()
        tmp_fp.index = tmp_fp.index + 1
        savemat(file_name, {'PPG_fiducials': tmp_fp.to_records(index=True)})

        for key in BM_keys:

            file_name = (relative_path+tmp_dir+os.sep+temp_dirs[1]+os.sep+'%s_btwn_%s-%s.mat')%(s.name+'_'+key,s.start_sig,s.end_sig)
            file_names[key+'_vals_mat']=file_name
            tmp_df_vals=bm.bm_vals[key]
            matlab_struct = tmp_df_vals.to_dict(orient='list')
            savemat(file_name, {'PPG_vals': tmp_df_vals.to_records(index=True)})

            if len(bm.bm_stats)>0:
                file_name = (relative_path+tmp_dir+os.sep+temp_dirs[2]+os.sep+'%s_btwn_%s-%s.mat')%(s.name+'_'+key,s.start_sig,s.end_sig)
                file_names[key + '_stats_mat']=file_name
                tmp_df_stat=bm.bm_stats[key]
                savemat(file_name, {'PPG_stats': tmp_df_stat.to_records(index=True)})

            file_name = (relative_path+tmp_dir+os.sep+temp_dirs[3]+os.sep+'%s_btwn_%s-%s.mat')%(s.name+'_'+key,s.start_sig,s.end_sig)
            file_names[key + '_defs_mat']=file_name
            tmp_df_defs=bm.bm_defs[key]
            matlab_struct = tmp_df_defs.to_dict(orient='list')
            savemat(file_name, {'PPG_defs': tmp_df_defs.to_records(index=True)})

            file_name = (relative_path+tmp_dir+os.sep+temp_dirs[5]+os.sep+'%s_btwn_%s-%s.mat')%(s.name+'_'+key,s.start_sig,s.end_sig)
            file_names[key + '_defs_stats_mat']=file_name
            tmp_df_defs2 = tmp_df_defs.drop('name', axis=1)
            tmp_df_defs2.index =tmp_df_defs['name']
            tmp_df_stat2=tmp_df_stat.transpose()
            tmp_df_defs_and_stats=pd.concat([tmp_df_defs2, tmp_df_stat2], axis=1)
            savemat(file_name, {key: tmp_df_defs_and_stats.to_records(index=True)})

    if savingformat != "csv" and savingformat != "mat" and savingformat != "both":
        raise('The file format is not suported for data saving! You can use "mat" or "csv" file formats.')

    if print_flag: print('Results have been saved into the "'+tmp_dir+'".')

    return file_names


def load_fiducials(saved_fiducials=""):
    """
    :param saved_fiducials: path of the matlab struct of the saved fiducial points, where the name field is 'PPG_fiducials'
    :type saved_fiducials: str

    :return: fiducial points: DataFrame where the key is the name of the fiducial pints and the value is the list of fiducial points
    """
    try:
        loaded_fp = scipy.io.loadmat(saved_fiducials)['PPG_fiducials']
        python_dict = {}
        for field in loaded_fp.dtype.names:
            python_dict[field] = np.squeeze(np.squeeze(loaded_fp[field]))

        fiducials = pd.DataFrame(python_dict)
        for fp_key in python_dict.keys():
            fiducials[fp_key] = [x[0, 0] for x in fiducials[fp_key]]

        return fiducials
    except:
        print('Invalid field name. The supported MATLAB field name for fiducial points is "PPG_fiducials".')
        return




import pyPPG
from pyPPG import PPG, Fiducials, Biomarkers
from pyPPG.datahandling import load_data, plot_fiducials, save_data, load_fiducials
import pyPPG.preproc as PP
import pyPPG.fiducials as FP
import pyPPG.biomarkers as BM
import pyPPG.ppg_sqi as SQI

import numpy as np
import sys
import json
import pandas as pd
import scipy.io

###########################################################################
################################## EXAMPLE ################################
###########################################################################
def ppg_example(data_path="", fs=0, start_sig=0, end_sig=-1, fiducials=pd.DataFrame(), process_type="both", channel="Pleth",
                filtering=True, fL=0.5000001, fH=12, order=4, sm_wins={'ppg':50,'vpg':10,'apg':10,'jpg':10}, correction=pd.DataFrame(),
                savingfolder="temp_dir", savefig=True, show_fig=True, savingformat="both", print_flag=True, use_tk=False, check_ppg_len=True,
                saved_fiducials="", savedata=True):
    '''
    This is an example code for PPG analysis. The main parts:
        1) Loading a raw PPG signal: various file formats such as .mat, .csv, .txt, or .edf.
        2) Get Fiducial points: extract the fiducial points of PPG, PPG', PPG'' and PPG'" signals
        3) Plot Fiducial Points
        4) Get Biomarkers: extract 74 PPG biomarkers in four categories:
            - PPG signal
            - Signal ratios
            - PPG derivatives
            - Derivatives ratios
        5) Get Statistics: summary of the 74 PPG biomarkers
        6) SQI calculation: calculates the PPG Signal Quality Index
        7) Save data: save the extracted Fiducial points, Biomarkers, and Statistics into .csv file

    :param data_path: path of the PPG signal
    :type data_path: str
    :param fs: sampling_frequency
    :type fs: int
    :param start_sig: beginning the of signal in sample
    :type start_sig: int
    :param end_sig: end of the signal in sample
    :type end_sig: int
    :param fiducials: DataFrame of the fiducial points
    :type fiducials: pyPPG.Fiducials DataFrame
    :param process_type: the type of the process, which can be "fiducials", "biomarkers", or "both"
    :type process_type: str
    :param channel: channel of the .edf file
    :type channel: channel of the .edf file
    :param filtering: a bool for filtering
    :type filtering: bool
    :param fL: Lower cutoff frequency (Hz)
    :type fL: float
    :param fH: Upper cutoff frequency (Hz)
    :type fH: float
    :param order: Filter order
    :type order: int
    :param sm_wins: dictionary of smoothing windows in millisecond:
        - ppg: window for PPG signal
        - vpg: window for PPG' signal
        - apg: window for PPG" signal
        - jpg: window for PPG'" signal
    :type sm_wins: dict
    :param correction: DataFrame where the key is the name of the fiducial points and the value is bool
    :type correction: DataFrame
    :param savingfolder: location of the saved data
    :type savingfolder: str
    :param savefig: a bool for current figure saving
    :type savefig: bool
    :param show_fig: a bool for show figure
    :type show_fig: bool
    :param savingformat: file format of the saved date, the provided file formats .mat, .csv, or both
    :type savingformat: str
    :param print_flag: a bool for print message
    :type print_flag: bool
    :param use_tk: a bool for using tkinter interface
    :type use_tk: bool
    :param check_ppg: a bool for checking ppg length and sampling frequency
    :type check_ppg: bool
    :param saved_fiducials: path of the file of the saved fiducial points
    :type saved_fiducials: str
    :param savedata: a bool for saving data
    :type savedata: bool

    :return: file_names: dictionary of the saved file names

    Example:

        .. code-block:: python

            from pyPPG.example import ppg_example

            # run example code
            ppg_example()

    '''

    ## Loading a raw PPG signal
    signal = load_data(data_path=data_path, fs=fs, start_sig=start_sig, end_sig=end_sig, channel=channel, use_tk=True, print_flag=print_flag)

    ## Preprocessing
    # Initialise the filters
    prep = PP.Preprocess(fL=fL, fH=fH, order=order, sm_wins=sm_wins)

    # Filter and calculate the PPG, PPG', PPG", and PPG'" signals
    signal.filtering = filtering
    signal.fL = fL
    signal.fH = fH
    signal.order = order
    signal.sm_wins = sm_wins
    signal.ppg, signal.vpg, signal.apg, signal.jpg = prep.get_signals(s=signal)

    # Initialise the correction for fiducial points
    corr_on = ['on', 'dn', 'dp', 'v', 'w', 'f']
    correction.loc[0, corr_on] = True
    signal.correction=correction

    ## Create a PPG class
    s = PPG(s=signal, check_ppg_len=check_ppg_len)

    ## Get Fiducial points
    if process_type == 'fiducials' or process_type == 'both':
        # Initialise the fiducials package
        fpex = FP.FpCollection(s=s)

        # Extract fiducial points
        fiducials = fpex.get_fiducials(s=s)
        if print_flag: print("Fiducial points:\n", fiducials + s.start_sig)

        # Create a fiducials class
        fp = Fiducials(fp=fiducials)

        # Save data
        if savedata:
            fp_new = Fiducials(fp=fp.get_fp() + s.start_sig)
            file_names=save_data(savingformat=savingformat, savingfolder=savingfolder, print_flag=print_flag, s=s, fp=fp_new)

    ## PPG SQI

        # Calculate SQI
        ppgSQI = round(np.mean(SQI.get_ppgSQI(ppg=s.ppg, fs=s.fs, annotation=fp.sp)) * 100, 2)
        if print_flag: print('Mean PPG SQI: ', ppgSQI, '%')

    ## Plot fiducial points
        plot_fiducials(s=s, fp=fp, savefig=savefig, savingfolder=savingfolder, show_fig=show_fig, print_flag=print_flag, use_tk=use_tk)

    ## Load saved fiducial points from MATLAB struct
    if ".mat" in saved_fiducials:
        tmp_fp1 = load_fiducials(saved_fiducials=saved_fiducials)
        tmp_fp2 = tmp_fp1[(tmp_fp1['on']>= s.start_sig) & (tmp_fp1['off']<= s.end_sig)]
        fiducials = tmp_fp2-s.start_sig
        fiducials.index =range(0,len(fiducials))

        ## Get Biomarkers and Statistics
    if (process_type == 'biomarkers' or process_type == 'both') and len(fiducials)>0:
        # Initialise the biomarkers package
        fp = Fiducials(fp=fiducials)

        bmex = BM.BmCollection(s=s, fp=fp)

        # Extract biomarkers
        bm_defs, bm_vals, bm_stats = bmex.get_biomarkers()

        if print_flag:
            tmp_keys = bm_stats.keys()
            print('Statistics of the biomarkers:')
            for i in tmp_keys: print(i, '\n', bm_stats[i])

        # Create a biomarkers class
        bm = Biomarkers(bm_defs=bm_defs, bm_vals=bm_vals, bm_stats=bm_stats)

        # Save data
        if savedata:
            fp_new = Fiducials(fp=fp.get_fp() + s.start_sig)
            file_names=save_data(savingformat=savingformat, savingfolder=savingfolder, print_flag=print_flag, s=s, fp=fp_new, bm=bm)

    if print_flag: print('Program finished')

    return file_names


###########################################################################
############################## RUN EXAMPLE CODE ###########################
###########################################################################
if __name__ == "__main__":

    if len(sys.argv) > 1:
        input_data = json.loads(sys.argv[1])
        function_name = input_data['function']
        function_args = input_data['args']

        if function_name == 'ppg_example':
            file_names = ppg_example(**function_args)
            print(json.dumps(file_names))

        else:
            print("Invalid function name")
    else:
        print("Please provide function name and arguments as JSON string")
        ppg_example(savefig=True)


from pyPPG.pack_ppg._ErrorHandler import _check_shape_, WrongParameter
import pandas as pd

class PPG:
    '''
    This is class for the input PPG parameters.
    '''
    def __init__(self,s={}, check_ppg_len=True):
        """
        :param s: dictionary of the PPG signal:

            * s.start_sig: beginning of the signal in sample
            * s.end_sig: end of the signal in sample
            * s.v: a vector of PPG values
            * s.fs: the sampling frequency of the PPG in Hz
            * s.name: name of the record
            * s.v: 1-d array, a vector of raw PPG values
            * s.fs: the sampling frequency of the PPG in Hz
            * s.ppg: 1-d array, a vector of the filtered PPG values
            * s.vpg: 1-d array, a vector of the filtered PPG' values
            * s.apg: 1-d array, a vector of the filtered PPG" values
            * s.jpg: 1-d array, a vector of the filtered PPG'" values
            * s.filtering: a bool for filtering
            * s.fL: Lower cutoff frequency (Hz)
            * s.fH: Upper cutoff frequency (Hz)
            * s.order: Filter order
            * s.sm_wins: dictionary of smoothing windows in millisecond:
                - ppg: window for PPG signal
                - vpg: window for PPG' signal
                - apg: window for PPG" signal
                - jpg: window for PPG'" signal
            * s.correction: DataFrame where the key is the name of the fiducial points and the value is bool
        :type s: DotMap
        :param check_ppg_len: a bool for checking ppg length and sampling frequency
        :type check_ppg_len: bool

        """

        if s.fs <= 0:
            raise WrongParameter("Sampling frequency should be strictly positive")

        if check_ppg_len: _check_shape_(s.v, s.fs)

        s.check_ppg_len=check_ppg_len

        try:
            s.start_sig>0
        except:
            s.start_sig = 0

        try:
            s.end_sig>-1
        except:
            s.end_sig = -1

        # Initialise the correction for fiducial points
        if len(s.correction)<1:
            corr_on = ['on', 'dn', 'dp', 'v', 'w', 'f']
            corr_off = ['dn']
            correction=pd.DataFrame()
            correction.loc[0, corr_on] = True
            correction.loc[0, corr_off] = False
            s.correction=correction

        keys=s.keys()
        keys_list = list(keys)
        for i in keys_list:
            exec('self.'+i+' = s[i]')

    def get_s(self):
        """
        This function retrieves the dictionary of the PPG signal.

        :return: s: dictionary of the PPG signal
        """
        keys = self.__dict__.keys()
        keys_list = list(keys)
        s = {}
        for i in keys_list:
            s[i] = getattr(self, i, None)

        return pd.DataFrame(s)


class Fiducials:
        '''
        This is class for the PPG fiducial points.
        '''
        def __init__(self, fp: pd.DataFrame):
            """
            :param fiducials: DataFrame where the key is the name of the fiducial points and the value is the list of fiducial points PPG Fiducials Points

                * PPG signal (fp.on, fp.sp, fp.dn, fp.dp): List of pulse onset, systolic peak, dicrotic notch, diastolic peak
                * 1st derivative (fp.u, fp.v, fp.w): List of points of 1st maximum and minimum in 1st derivitive between the onset to onset intervals
                * 2nd derivative (fp.a, fp.b, fp.c, fp.d, fp.e, fp.f): List of maximum and minimum points in 2nd derivitive between the onset to onset intervals
                * 3rd derivative (fp.p1, fp.p2): List of points of 1st maximum and minimum in 3rd derivitive between the onset to onset intervals
            :type fiducials: DataFrame

            """

            keys = fp.keys()
            keys_list = list(keys)
            for i in keys_list:
                exec('self.' + i + ' = fp[i]')

        def get_fp(self):
            """
            This function retrieves the struct of the fiducial points.

            :return: fp: DataFrame of the fiducial points
            """

            keys = self.__dict__.keys()
            keys_list = list(keys)
            fp={}
            for i in keys_list:
                fp[i]=getattr(self, i, None)

            return pd.DataFrame(fp)

        def get_row(self, row_index: int):
            """
            This function retrieves the specified row from the DataFrame of fiducial points.

            :param row_index: the index corresponding to the row in the fiducial points DataFrame
            :type row_index: int

            :return: the corresponding row in the fiducial points DataFrame
            """
            keys = self.__dict__.keys()
            keys_list = list(keys)
            fp_row={}
            for i in keys_list:
                fp_row[i]=getattr(self, i, None)[row_index]

            return pd.DataFrame(fp_row,[row_index])

class Biomarkers:
        '''
        This is class for the PPG biomarkers.
        '''
        def __init__(self, bm_defs={}, bm_vals={}, bm_stats={}):
            """
                This class constitutes a comprehensive dictionary encompassing biomarker definitions, values, and statistics. Each dictionary is organized into the subsequent subdirectories:
                    * ppg_sig: description for the PPG signal
                    * sig_ratios: description for the Signal ratios
                    * ppg_derivs: description for the PPG derivatives
                    * derivs_ratios: description for the Derivatives ratios

                :param bm_defs: dictionary with name, definition and unit of biomarkers in different categories:
                :type bm_defs: dict
                :param bm_vals: dictionary with values of biomarkers in different categories:
                :type bm_vals: dict
                :param bm_stats: data frame with summary of PPG features
                :type bm_stats: dict

            """

            for i in ['bm_defs', 'bm_vals', 'bm_stats']:
                exec('self.' + i + ' = ' +i)

        def get_bm(self):
            """
            This function retrieves the dictionary of the biomarkers.

            :return: bm: dictionary of the biomarkers
            """

            keys = self.__dict__.keys()
            keys_list = list(keys)
            bm={}
            for i in keys_list:
                bm[i]=getattr(self, i, None)

            return pd.DataFrame(bm)



from setuptools import setup, find_packages
import io
import os

with io.open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()


with open('docs'+os.sep+'requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='pyPPG',
    version='1.0.47',
    description='pyPPG: a python toolbox for PPG morphological analysis.',
    author='Marton A. Goda, PhD; Peter H. Charlton, PhD',
    author_email="marton.goda@campus.technion.ac.il",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/godamartonaron/GODA_pyPPG",
    project_urls={"Bug Tracker": "https://github.com/pypa/sampleproject/issues",},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    packages={"pyPPG", "pyPPG"+os.sep+"ppg_bm", "pyPPG"+os.sep+"pack_ppg", "pyPPG"+os.sep+"validation"},
    package_data={
        "pyPPG": ["*"],
        "pyPPG"+os.sep+"ppg_bm": ["*"],
        "pyPPG"+os.sep+"pack_ppg": ["*"],
        "pyPPG"+os.sep+"validation": ["pw_anal.py, __init__.py"],
    },

    install_requires=[required],
    python_requires=">=3.10",
    include_package_data=True,
)